package com.perisic.jshadow;

import java.lang.reflect.*;
import java.util.*;

/**
* A common base class for all shadowable objects. 
* Basically adding a shadow  to an 
* object <code>obj</code> means that a method call to <code>obj</code> will be executed
* in the shadow instead in <code>obj</code>. If the shadow does not define the method
* it will be passed to <code>obj</code>. If <code>obj</code> is shadowed by more than
* one shadow the most recently added shadow will be tried first.
* <ul>
* <li> An objects can only be shadowed if it inherits Shadowable (this class).
* <li> A method must contain only objects as parameters or return values (this will 
*      become a minor problem when autoboxing is available in jikes) The number of
*      parameters of a method must be not more than six. 
* <li> A shadow is an object that inherits {@link Shadow}
* <li> A shadow cannot be shadowed, but an object can have multiple shadows. 
* <li> A class can be shadowed. That means that all objects that are instantiated
*      from this class will be shadowed. 
* </ul> 
* This package partly mimics the behavior of shadows known from LPC, see for example
* <a href="http://genesis.cs.chalmers.se/coding/lpcdoc/lpc_58.html" target="_blank">http://genesis.cs.chalmers.se/coding/lpcdoc/lpc_58.html</a>
* 
*<UL>
*<li> First version: 16.09.2004
*<li>
* Copyright: (c) Marc Conrad, 2004
*<li>
* See http://perisic.com/shadow for information about this package. 
*<li>The com.perisic.shadow library is distributed under the terms of the 
* <a href="http://www.gnu.org/copyleft/lesser.html" target="_blank">
GNU Lesser General Public License (LGPL)</a>. 
*<li> If you require the package under a different licence please contact me.
*<li> disclaimer: The classes are provided "as is".
* There is no warranty implied by using the com.perisic.shadow package.
* <li> Note: This class needs a special adapted version of the <a name="jikes">jikes</a> compiler (available on
* <a href="http://perisic.com/shadow">http://perisic.com/shadow</a>) 
* for successful compilation. Compile this 
*      class with the command jikes -dy
* </UL>
*/

public abstract class Shadowable { 
/**
*     Holds all Shadows that are added to an object.
*/
      private Vector shadows = new Vector(); 
/** 
*     For each class that is shadowed it holds a vector of classes that 
*     shadow that class. 
*/

        private static Hashtable classShadows = new Hashtable(); 
/**
*     When constructed the instances of the class shadows of this class and all 
*     parent classes 
*     are added to this object. 
*/
      public Shadowable() {       
        for( Class c = this.getClass(); c != Shadowable.class; c =  c.getSuperclass() ) {
            Vector vc = getClassShadows(c);
          if( vc != null ) {
           for(int j = 0; j < vc.size(); j++ ) { 
        Class cs = (Class) vc.get(j);
            try { 
              addShadow((Shadow) cs.newInstance()); 
            } catch ( InstantiationException e ) { 
                System.out.println("Shadowable: "+e);  
                e.printStackTrace();
            } catch (IllegalAccessException e) {
                    System.out.println("Shadowable: "+e);
                    e.printStackTrace();
                } 
            }
          }
        }
      }
/**
*     Add a shadow to this object. A shadow receives all method calls
*     sent to this object. If the shadow defines the method
*     it will be processed in the shadow.  
*
*     @param shadow The shadow that is added to this object.
*    
*     
*/
      public void addShadow(Shadow shadow) { 
            shadows.add(shadow); 
            shadow.initShadow(this); 
            }
            
/**
*    An instance of  <code>shadow</code> is added whenever an <code>owner</code>object
*    is instantiated. The order in that the shadows
*    are added to the object is the same order as the shadow classes are
*    added to <code>owner</code>. 
*       
*    @param owner The class that is to be shadowed
*    @param shadow The class that shadows <code>owner</code> 
*     <!-- -->
*     (Defined only on package level, the public method is 
*      available in com.perisic.jshadow.Util.)     
*/     
  
     static void addShadow(Class owner, Class shadow) {
            Vector vc = (Vector) classShadows.get(owner);
            if( vc == null ) {
                vc = new Vector();
                }
            vc.add(shadow);
            classShadows.put(owner,vc);
            }
/**
*     Removes the class <code>shadow</code> that has been added to the class
*     <code>owner</code>. Note: the shadow <em>objects</em> that have been 
*     already added to instances of <code>owner</code> are not removed.
*     To remove shadows from objects use {@link #removeShadow(Shadow)}
* 
*     @param owner the class that is shadowed
*     @param shadow the shadow to be removed from the class
*     @return true if  <code>owner</code> has been shadowed by <code>shadow</code>
*     (Defined only on package level, the public method is 
*      available in com.perisic.jshadow.Util.)     
*    
*/            
      static boolean removeShadow(Class owner, Class shadow) {
            Vector vc = (Vector) classShadows.get(owner);
            if( vc == null ) { return false; }
            boolean result = vc.remove(shadow);
            if( vc.size() == 0 ) {
                classShadows.remove(owner);
                }
            return result;
            }
/**
*    Gets a vector of all shadows that belong to the class <code>owner</code>. 
*    This may be used to manipulate the class shadows manually by changing the
*    the vector.
*    @param owner the class whose shadows are returned
*    @return all classes added to <code>owner</code>  
*     (Defined only on package level, the public method is 
*      available in com.perisic.jshadow.Util.)     
*/
            
     static Vector getClassShadows(Class owner) {
            return (Vector)  classShadows.get(owner);
            }
     
/**
*     Gets all classes that have been shadowed.
*     @return an array of all classes that are shadowed.
*     (Defined only on package level, the public method is 
*      available in com.perisic.jshadow.Util.)       
*/        
     static Class [] allShadowedClasses() {
            return (Class []) ((Vector) classShadows.keySet()).toArray(new Class[0]);
            }
/**
*     Removes a shadow from this object. 
*     @return true if this object has been shadowed by <code>shadow</code>
*     <!-- / --> 
*/            
      public boolean removeShadow(Shadow shadow) { 
            return shadows.remove(shadow); 
            }
            
/**
*     Gets all shadows that shadow this object. Modification of the
*     returned vector will modify the shadows. 
*     @return a vector of all shadows of this object
*/
      public Vector getShadows() { 
            return shadows;
            }

/**
*     Every method call with six arguments to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method. 
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler. 
*/
      public Object receive_call( Object whodunnit, Object methodName, 
                 Object ob1, Object ob2, Object ob3, Object ob4,
                 Object ob5, Object ob6 ) {
         return gen_receive_call( whodunnit, (String) methodName, 
                  new Object [] { ob1, ob2, ob3, ob4, ob5, ob6 } );
        }
/**
*     Every method call with five arguments to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method.
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler.  
*/
      public Object receive_call( Object whodunnit, Object methodName, 
                 Object ob1, Object ob2, Object ob3, Object ob4,
                 Object ob5 ) {
         return gen_receive_call( whodunnit, (String) methodName, 
                  new Object [] { ob1, ob2, ob3, ob4, ob5 } );
        }
/**
*     Every method call with four arguments to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method. 
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler. 
*/
      public Object receive_call( Object whodunnit, Object methodName, 
                 Object ob1, Object ob2, Object ob3, Object ob4 ) {
         return gen_receive_call( whodunnit, (String) methodName, 
                  new Object [] { ob1, ob2, ob3, ob4 } );
        }  
/**
*     Every method call with three arguments to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method. 
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler. 
*/
      public Object receive_call( Object whodunnit, Object methodName, 
                 Object ob1, Object ob2, Object ob3 ) {
         return gen_receive_call( whodunnit, (String) methodName, 
                new Object [] { ob1, ob2, ob3 } );
        }
        
/**
*     Every method call with two arguments to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method. 
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler. 
*/
      public Object receive_call( Object whodunnit, Object methodName, 
                       Object ob1, Object ob2 ) {
         return gen_receive_call( whodunnit, (String) methodName, 
                       new Object [] { ob1, ob2 } );
        }
      
/**
*     Every method call with one argument to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method. 
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler. 
*/
      public Object receive_call( Object whodunnit, Object methodName, 
                       Object ob1 ) {
         return gen_receive_call( whodunnit, (String) methodName, 
                       new Object [] { ob1 } );
        }
/**
*     Every method call with no arguments to this object is 
*     received by this method. The first argument is the caller
*     of the method. The second argument is the name of the method. 
*     This method is usually not called explicitely but will be 
*     implicitely called by the Java virtual machine if the code
*     has been with the <a href="#jikes">dynamic extension of the jikes</a> compiler. 
*/
      public Object receive_call( Object whodunnit, Object methodName 
                        ) {
        return gen_receive_call( whodunnit, (String) methodName, 
                       new Object [] { } );
        }
/**
*      Executes the method <code>methodName</code> with the arguments <code>args</code>.
*      When the method is shadowed (that means when the object has a <code>Shadow</code> 
*      object added and the <code>Shadow</code> object defines the method), 
*      then the method defined
*      in the shadow object is executed instead of the method in this object. 
*      <p>
*      If an object has more than one shadow added, the last shadow added to the method
*      is checked first.  
*      <p>
*      Multiple shadows on the same objects define a hierarchy 
*      <em>S<sub>1</sub></em>, <em>S<sub>2</sub></em>, <em>S<sub>3</sub></em>, ... of shadows that is defined by the 
*      order in that these shadows are added. If a method is called by a shadow 
*      <em>S<sub>i</sub></em>, only the shadows <em>S<sub>j</sub></em> with <em>j &lt; i</em> are checked. 
*      <p>
*      If the method is not found in the shadows and in the object itself a 
*      {@link ShadowException} with a <code>NoSuchMethodException</code> as cause
*      is thrown. 
*      <p> 
*      It the method is found but throws an exception, a <code>ShadowException</code>
*      is thrown where the cause is the exception originally thrown by the method.
*/
       
            
       Object gen_receive_call(Object whodunnit, String methodName, Object [] args ) {  
                        
           // If the call comes from a shadow, we start searching from that index, i.e
           // when the call comes from the second shadow we do not look in shadow three or
           // two, but in shadow one.
           int n = 0; 
          while( n < shadows.size() && whodunnit != shadows.get(n) ) { n = n+1; }
           // n is now equals shadows.size() or the index of whodunnit in shadows.
           // The first shadow to check for the function is n.  
           Object s; 
              // The for loop assigns s to all shadows and finally to this. 
              // The most recent added shadow is checked first.
           for( int i = n; i >= 0; i-- ) {
                if ( i == 0 ) { 
                    s = this; 
                    }
                else {
                    s = shadows.get(i-1);
                    }
                 try { 
                  Class cs = s.getClass(); 
                  Vector methods 
                     = Util.getMostSpecificMethod(cs, methodName, args );
          if( methods.size() > 1 ) {
            throw new ShadowException("Ambigous Method Invocation");
            }
          else if( methods.size() == 1 ) {
                    Method mcs = (Method) methods.get(0);
                   return mcs.invoke(s, args);
          

            }
          else { // methods.size() == 0 
                    if( i == 0 ) { 
                        // The method is not defined in the class and not
                        // in the shadows. 
            throw new ShadowException("No such Method defined");
               }
            }
                 } catch( IllegalAccessException e ) { 
                       // ignore, try next shadow
                 } catch (IllegalArgumentException e ) { 
                   // This should not happen as the method has as parameter types
                   // the classes of the objects. Otherwise it will not be invoked.
                   throw new ShadowException(e);
                 } catch (InvocationTargetException e) { 
                 // The method has been found and thrown an exception.
                    throw new ShadowException(e);
                 }
           }
       throw new ShadowException("Unreachable statement reached.");
       }
}
